from __future__ import division, print_function

import pytest
import pickle
import io
import numpy as np
import random
from numpy.testing import assert_allclose

from ...neural.optimizers import SGD
from ...neural._classes.affine import Affine

np.random.seed(2)
random.seed(0)


@pytest.fixture
def model():
    model = Affine(2, 2)
    return model


def test_init(model):
    assert model.nO == 2
    assert model.nI == 2
    assert model.W is not None
    assert model.b is not None


def test_predict_bias(model):
    input_ = model.ops.allocate((1, model.nI,))
    target_scores = model.ops.allocate((1, model.nO))
    scores = model(input_)
    assert_allclose(scores[0], target_scores[0])

    # Set bias for class 0
    model.b[0] = 2.
    target_scores[0, 0] = 2.
    scores = model(input_)
    assert_allclose(scores, target_scores)

    # Set bias for class 1
    model.b[1] = 5.
    target_scores[0, 1] = 5.
    scores = model(input_)
    assert_allclose(scores, target_scores)


@pytest.mark.parametrize(
    'X,expected', [
        (np.asarray([0,0]), [0., 0.]),
        (np.asarray([1,0]), [1., 0.]),
        (np.asarray([0,1]), [0., 1.]),
        (np.asarray([1,1]), [1., 1.])
    ])
def test_predict_weights(X, expected):
    W = np.asarray([1.,0.,0.,1.]).reshape((2, 2))
    bias = np.asarray([0.,0.])

    model = Affine(W.shape[0], W.shape[1])
    model.W[:] = W
    model.b[:] = bias

    scores = model.predict_one(X)
    assert_allclose(scores, expected)



def test_update():
    W = np.asarray([1.,0.,0.,1.]).reshape((2, 2))
    bias = np.asarray([0.,0.])

    model = Affine(2, 2)
    model.W[:] = W
    model.b[:] = bias
    sgd = SGD(model.ops, 1.0, L2=0.0)
    sgd.averages = None
    
    ff = np.asarray([[0,0]])
    tf = np.asarray([[1,0]])
    ft = np.asarray([[0,1]])
    tt = np.asarray([[1,1]])

    # ff, i.e. 0, 0
    scores, finish_update = model.begin_update(ff)
    assert_allclose(scores[0,0], scores[0,1])
    # Tell it the answer was 'f'
    gradient = np.asarray([[-1., 0.]])
    finish_update(gradient, sgd)

    assert model.b[0] == 1.
    assert model.b[1] == 0.
    # Unchanged -- input was zeros, so can't get gradient for weights.
    assert model.W[0, 0] == 1.
    assert model.W[0, 1] == 0.
    assert model.W[1, 0] == 0.
    assert model.W[1, 1] == 1.

    # tf, i.e. 1, 0
    scores, finish_update = model.begin_update(tf)
    # Tell it the answer was 'T'
    gradient = np.asarray([[0., -1.]])
    finish_update(gradient, sgd)

    assert model.b[0] == 1.
    assert model.b[1] == 1.
    # Gradient for weights should have been outer(gradient, input)
    # so outer([0, -1.], [1., 0.])
    # =  [[0., 0.], [-1., 0.]]
    assert model.W[0, 0] == 1. - 0.
    assert model.W[0, 1] == 0. - 0.
    assert model.W[1, 0] == 0. - -1.
    assert model.W[1, 1] == 1. - 0.



from .model import Model


class Concatenated(Model):
    def __init__(self, layer):
        Model.__init__(self)
        self._layers.append(layer)
        self.on_data_hooks.append(on_data)

    @property
    def nO(self):
        return self._layers[0].nO + self._layers[0].nI

    @property
    def nI(self):
        return self._layers[0].nI

    def __call__(self, X):
        Y = self._layers[0](X)
        if isinstance(X, list) or isinstance(X, tuple):
            return [self.ops.xp.hstack([Y[i], X[i]]) for i in range(len(X))]
        else:
            return self.ops.xp.hstack([Y, X])

    def begin_update(self, X, drop=0.):
        y, bp_y = self._layers[0].begin_update(X, drop=drop)
        if isinstance(X, list) or isinstance(X, tuple):
            output = [self.ops.xp.hstack((y[i], X[i])) for i in range(len(X))]
            nr_y = y[0].shape[1]
        else:
            output = self.ops.xp.hstack((y, X))
            nr_y = y.shape[1]
        def concatenated_bwd(d_output, sgd=None):
            if isinstance(d_output, list) or isinstance(d_output, tuple):
                dY = []
                for i in range(len(d_output)):
                    dY.append(self.ops.xp.ascontiguousarray(d_output[i][:, :nr_y]))
                dX = bp_y(dY, sgd=sgd)
                for i in range(len(d_output)):
                    dX[i] += d_output[i][:, nr_y:]
            else:
                dY = self.op.xp.ascontiguousarray(d_output[:, :nr_y])
                dX = bp_y(dY, sgd=sgd)
                dX += d_output[:, nr_y:]
            return dX
        return output, concatenated_bwd


def on_data(self, X, y=None):
    for layer in self._layers:
        for hook in layer.on_data_hooks:
            hook(layer, X, y)
        if hasattr(layer, 'W'):
            layer.W.fill(0)
